from operator import itemgetter

from cumin.objectselector import ObjectSelector, ObjectLinkColumn, ObjectTable, \
    ObjectTableColumn, ObjectQmfTable, ObjectQmfSelectorTable, ObjectSelectorTaskForm, ObjectSelectorTask
from cumin.objectframe import ObjectFrame, ObjectTaskForm, ObjectDetails
from cumin.util import *
from cumin.qmfadapter import ObjectQmfAdapter
from cumin.task import Task, TaskLink

from datetime import datetime

from sage.util import MethodResult

from wooly import Session, Widget
from wooly.datatable import *
from wooly.table import LinkColumnCell
from wooly.util import StringCatalog, xml_escape

from rosemary.model import RosemaryObject

strings = StringCatalog(__file__)

def compareHadoopIds(x, y):
    return int(float(x.id) - float(y.id))

class HadoopAdapter(ObjectQmfAdapter):
    def __init__(self, app, cls, check_owner=False):
        super(HadoopAdapter, self).__init__(app, cls)
        self.cls = cls
        self.check_owner = check_owner

    def get_count(self, values):
        data = self.do_get_data(values)
        return len(data)

    def get_sage_results(self, values):
        raise Exception("Not implemented")

    def do_get_data(self, values):
        (status, data) = self.get_sage_results(values)
        log.debug("Return status from sage call is: %s" % status)
        if data is None or len(data) == 0:
            return {}

        return data

    def process_record(self, key, record):
        field_data = list()
        for column in self.columns:
            try:
                if column.name.lower() in ["parent_id", "parent_ipc"]:
                    val = record.parent[column.name.lower().split("_")[1]]
                elif column.name == "State":
                    val = record[column.name.lower()].title()
                else:
                    val = record[column.name.lower()]
            except AttributeError:
                if (column.name == "Ipc"):
                    val = record.ref.ipc
                else:
                    val = record.ref.id
            field_data.append(val)
        return field_data

    def get_owner_filter(self, session):
        if self.check_owner:
            owner_list = [session.get_user().name]
        else:
            owner_list = None
        return owner_list

    def filter_rows(self, rows, values, options):
        rows = super(HadoopAdapter, self).filter_rows(rows, values, options)
        try:
            if getattr(options, "parent_filter") is not None:
                rows = self.filter_by_parent(rows, options.parent_filter)
        except AttributeError, e:
            pass
        return rows

    def filter_by_parent(self, rows, parent_filter):
        filtered_rows = []
        for row in rows:
            if row[parent_filter["column"].field.index] == parent_filter["value"]:
                filtered_rows.append(row)
        return filtered_rows

class HadoopNodeDeleteForm(ObjectSelectorTaskForm):
    def __init__(self, app, name, task, cls):
        super(HadoopNodeDeleteForm, self).__init__(app, name, task)
        self.task = task

        self.hadoophost = Parameter(app, "hadoophost")
        self.add_parameter(self.hadoophost)

    def get_selection(self, session):
        ids = self.ids.get(session)
        self.selection.set(session, ids)
        return len(ids)

    def render_form_class(self, session):
        return " ".join((super(HadoopNodeDeleteForm, self).render_form_class(session), "mform"))

    def render_content(self, session, *args):
        # need to show "content" which has the selected list from objectselectortaskform
        # also need to render any fields by hand here since
        # the usual content block gets swapped out
        content = ""
#        content += '''<table class="FormFieldSet"><tbody>'''
#        content += self.hadoophost.render(session)
#        content += '''</tbody></table>'''
        content += self.content.render(session, *args)
        return content

class HadoopNodeCreateForm(ObjectTaskForm):
    def render_form_class(self, session):
        return " ".join((super(HadoopNodeCreateForm, self).render_form_class(session), "mform"))


class HadoopHostField(ScalarField):
    def __init__(self, app, name):
        super(HadoopHostField, self).__init__(app, name, None)

        self.param = StringParameter(app, "param")
        self.add_parameter(self.param)

        self.input = self.HadoopHostOptions(app, "input", self.param)
        self.add_child(self.input)

    def get(self, session):
        return self.input.get(session)

    def validate(self, session):
        super(HadoopHostField, self).validate(session)

        name_nodes = self.input.get_items(session)

        if not name_nodes:
            error = FormError("There is no hadoop host to submit to")
            self.form.errors.add(session, error)

    def render_title(self, session):
        return "Hadoop host"

    class HadoopHostOptions(OptionInputSet):
        def do_process(self, session):
            id = self.param.get(session)

            if id is None:
                items = self.get_items(session)

                if items:
                    self.param.set(session, items[0])

            super(HadoopHostField.HadoopHostOptions, self).do_process(session)

        def do_get_items(self, session):
            hosts = self.app.remote.get_hosts("SCHEDULER", "HADOOP")
            return hosts

        def render_item_value(self, session, item):
            return item

        def render_item_content(self, session, item):
            return xml_escape(item)

        def render_item_selected_attr(self, session, item):
            if item == self.param.get(session):
                return "selected=\"selected\""

class BinFileLoc(StringField):
    def __init__(self, app, name):
        super(BinFileLoc, self).__init__(app, name)
        self.input.param.default = app.hadoop_bin_tarball

    def render_title(self, session):
        return "Location of hadoop bin-tar file"

class Owner(StringField):
        def render_title(self, session):
            return "Username of the owner"


class NameNodeField(ScalarField):
    def __init__(self, app, name):
        super(NameNodeField, self).__init__(app, name, None)

        self.param = IntegerParameter(app, "param")
        self.add_parameter(self.param)

        cls = app.model.com_redhat_cumin_grid_hadoop.NameNode

        self.input = self.NameNodeOptions(app, "input", self.param)
        self.add_child(self.input)

    def get(self, session):
        return self.input.get(session)

    def validate(self, session):
        super(NameNodeField, self).validate(session)

        name_nodes = self.input.get_items(session)

        if not name_nodes:
            error = FormError("There is no name node to bind to")
            self.form.errors.add(session, error)

    def render_title(self, session):
        return "Name node"

    class NameNodeOptions(OptionInputSet):
        def do_process(self, session):
            cls = self.app.model.com_redhat_cumin_grid_hadoop.NameNode
            id = self.param.get(session)

            if id is None:
                items = self.get_items(session)

                if items:
                    self.param.set(session, items[0])

            super(NameNodeField.NameNodeOptions, self).do_process \
                (session)

        def do_get_items(self, session):
            nodes = []
            (status, node_list) = self.app.remote.get_name_node_list()
            for node in node_list:
                nodes.append(node.ref)
            return sorted(nodes, cmp=compareHadoopIds)

        def render_item_value(self, session, item):
            value = None
            if item.ipc:
                value = item.ipc
            else:
                value = item.id
            return value

        def render_item_content(self, session, item):
            display_string = "%s (%s)" % (item.ipc, item.id)
            return xml_escape(display_string)

        def render_item_selected_attr(self, session, item):
            if item == self.param.get(session):
                return "selected=\"selected\""

class SubmittedColumn(ObjectTableColumn):
    def render_cell_content(self, session, data):
        d = super(SubmittedColumn, self).\
                         render_cell_content(session, data)
        datestring = datetime.fromtimestamp(d).strftime('%Y-%m-%d %H:%M:%S')
        return datestring

class UptimeColumn(ObjectTableColumn):
    def render_cell_content(self, session, record):
        value = self.field.get_content(session, record)
        days = value / 86400
        hours = (value / 3600) - (days * 24)
        minutes = (value / 60) - (days * 1440) - (hours * 60)
        return '%02d:%02d:%02d' % (days, hours, minutes)

class HadoopObject(RosemaryObject):
    def __init__(self, aviary_response, cls):
        self._class = cls
        self._save_time = datetime.now()
        properties = [prop for prop in cls._properties]
        for prop in properties:
            try:
                if prop.name.lower() in ["id", "ipc"]:
                    setattr(self, prop.name, aviary_response[1][0]["ref"][prop.name.lower()])
                elif prop.name.lower() in ["parent_id", "parent_ipc"]:
                    setattr(self, prop.name, aviary_response[1][0]["parent"][prop.name.lower().split("_")[1]])
                else:
                    setattr(self, prop.name, aviary_response[1][0][prop.name.lower()])
            except AttributeError, e:
                pass
        self.Name = self.Ipc
        self._id = self.Id

    def __repr__(self):
        name = self.__class__.__name__
        args = (name, self._class, self._id, self._save_time)
        return "%s(%s,%s,%s)" % args

class HadoopObjectDetails(ObjectDetails):
    def __init__(self, app, name, object):
        super(HadoopObjectDetails, self).__init__(app, name, object)

class HadoopNodeDeleteTask(ObjectSelectorTask):
    def enter(self, session):
        form_session = super(HadoopNodeDeleteTask, self).enter(session)
        self.form.hadoophost.set(form_session, self.module.frame.children_by_name[self.module.frame.name + ".ghschedframe"].host.get(session))
        return form_session

    def invoke(self, session, selection, *args):
        invoc = self.start(session, selection)

        try:
            self.do_invoke(invoc, selection, *args)
        except Exception, e:
            invoc.exception = e
            invoc.status = invoc.FAILED
            invoc.end()

class HadoopObjectLinkColumn(ObjectLinkColumn):
    def __init__(self, app, name, attr, id_attr, frame_path):
        super(HadoopObjectLinkColumn, self).__init__(app, name, attr, id_attr, frame_path)

        self.id_attr = id_attr
        self.frame_path = frame_path

        self.cell = HadoopLinkColumnCell(app, "cell")
        self.replace_child(self.cell)

class HadoopLinkColumnCell(LinkColumnCell):
    # needed so we can override display
    pass

class HadoopIpcColumn(ObjectTableColumn):
    def __init__(self, app, name, attr, width):
        super(HadoopIpcColumn, self).__init__(app, name, attr)
        self.width = width

    def render_cell_title(self, session, record):
        return self.field.get_content(session, record)

class HadoopHttpColumn(HadoopObjectLinkColumn):
    def render_cell_href(self, session, record):
        link = self.field.get_content(session, record)
        return link
