from cumin.objectselector import ObjectSelector, ObjectLinkColumn, ObjectTable, \
    ObjectTableColumn, ObjectQmfTable, ObjectQmfSelectorTable, ObjectSelectorTaskForm, ObjectSelectorTask
from cumin.objectframe import ObjectFrame, ObjectTaskForm
from cumin.util import *
from cumin.qmfadapter import ObjectQmfAdapter
from cumin.task import Task, TaskLink
from cumin.objectframe import ObjectDetails

from wooly.forms import StringField

from hadoop import *
from hadoopobjectframe import *

from tasktracker import TaskTrackerSelector, TaskTrackerFrame

from sage.util import *

from wooly.util import StringCatalog

strings = StringCatalog(__file__)

class JobTrackerAdapter(HadoopAdapter):
    def get_sage_results(self, values):
        owner_filter = self.get_owner_filter(values['session'])
        results = self.app.remote.get_job_tracker_list(owner=owner_filter)
        return results

class JobTrackerSelector(ObjectSelector):
    def __init__(self, app, name, module, check_owner=False):
        cls = app.model.com_redhat_cumin_grid_hadoop.JobTracker

        self.module = module

        super(JobTrackerSelector, self).__init__(app, name, cls)

        self.add_search_filter(self.table.id_col)
        self.table.adapter = JobTrackerAdapter(app, cls, check_owner)

        task = JobTrackerCreate(app, name + ".create", self.module)
        link = TaskLink(app, "jobTracker_create", task)
        self.links.add_child(link)

        JobTrackerDelete(app, self, "jobTracker_delete", self.module)

        self.enable_csv_export()

    def create_table(self, app, name, cls):
        return JobTrackerTable(app, name, cls)

    def render_title(self, session):
        return "Hadoop job trackers"

class JobTrackerTable(ObjectQmfSelectorTable):
    def __init__(self, app, name, cls):
        super(JobTrackerTable, self).__init__(app, name, cls)

        self.id_col = self.JobTrackerColumn(app, "idcol", cls.Id, cls.Id, ".jobtrackerframe")
        self.id_col.width = "10%"
        self.ipc_col = HadoopIpcColumn(app, "ipc", cls.Ipc, "20%")
        self.http_col = HadoopHttpColumn(app, "http", cls.Http, cls.Http, None)
        self.state_col = ObjectTableColumn(app, "statecol", cls.State)
        self.state_col.width = "10%"
        self.uptime_col = UptimeColumn(app, "uptimecol", cls.Uptime)
        self.uptime_col.width = "10%"
        self.owner_col = ObjectTableColumn(app, "ownercol", cls.Owner)

        self.add_column(self.id_col)
        self.add_column(self.ipc_col)
        self.add_column(self.http_col)
        self.add_column(self.state_col)
        self.add_column(self.uptime_col)
        self.add_column(self.owner_col)

    class JobTrackerColumn(ObjectLinkColumn):
        def render_cell_href(self, session, record):
            id = unescape_entity(record[self.id_field.index])
            frame = self.frame.children_by_name[self.frame.name + self.frame_path]
            return frame.get_href(session, id, frame.parent.host.get(session))

class JobTrackerDelete(HadoopNodeDeleteTask):
    def __init__(self, app, selector, name, module):
        super(JobTrackerDelete, self).__init__(app, selector)
        cls = app.model.com_redhat_cumin_grid_hadoop.JobTracker

        self.module = module

        self.form = JobTrackerDeleteForm(app, self.name, self, cls)
        self.invoc = None

    def do_invoke(self, invoc, ids, args):
        self.invoc = invoc
        (hadoophost) = args
        self.app.remote.stop_job_tracker(hadoophost, ids, invoc.make_callback())

    def get_title(self, session):
        return "Remove job trackers"

    def get_item_content(self, session, item):
        return xml_escape(item)


class JobTrackerDeleteForm(HadoopNodeDeleteForm):
    def process_submit(self, session):
        self.validate(session)
        ids = self.selection.get(session)
        hadoophost = self.hadoophost.get(session)
        url = self.return_url.get(session)
        self.page.redirect.set(session, url)

        if not self.errors.get(session):
            self.task.invoke(session, ids, (hadoophost))

    def render_title(self, session):
        return "Remove job trackers"

class JobTrackerCreate(Task):
    def __init__(self, app, name, module):
        super(JobTrackerCreate, self).__init__(app, name)
        cls = app.model.com_redhat_cumin_grid_hadoop.JobTracker

        self.module = module

        self.form = JobTrackerCreateForm(app, self.name, self, cls)
        self.invoc = None

    def callback(self, result):
        if result == False:
            self.invoc.status = self.invoc.FAILED
        self.invoc.end()

    def do_invoke(self, session, object, invoc, args):
        self.invoc = invoc
        (binfile, owner, hadoophost, name_node, description, count) = args

        try:
            self.app.remote.start_job_tracker(hadoophost, name_node, binfile, owner, description, count, invoc.make_callback())
        except Exception, e:
            invoc.status = invoc.FAILED
            log.debug("Creating job tracker failed", exc_info=True)
            invoc.end()

    def get_title(self, session, x):
        return "Create a job tracker"

    def enter(self, session, obj):
        form_session = wooly.Session(self.app.form_page)

        if obj:
            self.form.id.set(form_session, obj._id)

        self.form.hadoophost.set(form_session, self.app.gridhadoop.frame.children_by_name["gridhadoop.ghschedframe"].host.get(session))
        self.form.return_url.set(form_session, session.marshal())
        self.form.show(form_session)

        self.do_enter(session, obj, form_session)
        return form_session


class JobTrackerCreateForm(HadoopNodeCreateForm):
    def __init__(self, app, name, task, cls):
        super(JobTrackerCreateForm, self).__init__(app, name, task, cls)

        self.nameNode = NameNodeField(app, "namenode")
        self.add_field(self.nameNode)

        self.binfile = BinFileLoc(app, "binfile")
        self.add_field(self.binfile)

        self.hadoophost = Parameter(app, "hadoophost")
        self.add_parameter(self.hadoophost)

    def process_display(self, session):
        self.scheduler.validate(session)

    def process_submit(self, session):
        self.validate(session)

        url = self.return_url.get(session)
        self.page.redirect.set(session, url)

        if not self.errors.get(session):
            binfile = self.binfile.get(session)
            owner = session.client_session.username()
            hadoophost = self.hadoophost.get(session)
            name_node = self.nameNode.get(session)
            description = ""
            count = 1
            self.task.invoke(session, None, (binfile, owner, hadoophost, name_node, description, count))
            self.task.exit_with_redirect(session, url)

    def render_title(self, session):
        return "Create a job tracker"

class JobTrackerFrame(HadoopObjectFrame):
    def __init__(self, app, name, module):
        super(JobTrackerFrame, self).__init__(app, name, module)

        self.module = module

        self.view = JobTrackerObjectView(app, "view", self.object)
        self.replace_child(self.view)

        self.task_tracker_frame = TaskTrackerFrame(app, "tasktrackerframe", module)
        self.add_mode(self.task_tracker_frame)

        self.overview = JobTrackerOverview(app, self.name + ".jtoverview", self.object, self.module)
        self.view.add_tab(self.overview)

    def get_title(self, session):
        retval = ""
        try:
            retval = "Job tracker:'%s' (%s)" % (xml_escape(self.object.get(session).Name), xml_escape(self.object.get(session).Id))
        except Exception, e:
            pass
        return retval

    def get_object(self, session, id):
        host = self.frame.host.get(session)
        response = self.app.remote.get_job_tracker(self.frame.host.get(session), [id])
        obj = HadoopObject(response, self.app.model.com_redhat_cumin_grid_hadoop.JobTracker)
        return obj

class JobTrackerObjectView(HadoopObjectView):
    def add_details_tab(self):
        self.add_tab(JobTrackerDetails(self.app, "details", self.object))

class JobTrackerOverview(Widget):
    def __init__(self, app, name, node, module):
        super(JobTrackerOverview, self).__init__(app, name)
        self.add_child(TaskTrackerSelector(app, self.name + ".ttforjt", module, False))

    def render_title(self, session):
        return "Task trackers"

class JobTrackerDetails(HadoopObjectDetails):
    def __init__(self, app, name, object):
        super(JobTrackerDetails, self).__init__(app, name, object)
        self.object = object

    def render_title(self, session):
        return "Details"

    def render_name(self, session):
        '''sample....would only be called if JobTrackerDetails display was overridden to say {name}'''
        obj = self.object.get(session)
        return obj.Name
