from cumin.objectselector import ObjectSelector, ObjectLinkColumn, ObjectTable, \
    ObjectTableColumn, ObjectQmfTable, ObjectQmfSelectorTable, ObjectSelectorTaskForm
from cumin.objectframe import ObjectFrame, ObjectTaskForm
from cumin.util import *
from cumin.qmfadapter import ObjectQmfAdapter
from cumin.task import Task, TaskLink

from wooly.forms import StringField

from hadoop import *
from hadoopobjectframe import *
from datanode import DataNodeSelector, DataNodeFrame

from sage.util import *

from wooly.util import StringCatalog

strings = StringCatalog(__file__)

class NameNodeAdapter(HadoopAdapter):
    def get_sage_results(self, values):
        owner_filter = self.get_owner_filter(values['session'])
        results = self.app.remote.get_name_node_list(owner=owner_filter)
        return results

class NameNodeSelector(ObjectSelector):
    def __init__(self, app, name, module, check_owner=False):
        cls = app.model.com_redhat_cumin_grid_hadoop.NameNode

        super(NameNodeSelector, self).__init__(app, name, cls)

        self.add_search_filter(self.table.id_col)
        self.table.adapter = NameNodeAdapter(app, cls, check_owner)

        task = NameNodeCreate(app, name + ".create", module)
        link = TaskLink(app, "nameNode_create", task)
        self.links.add_child(link)

        NameNodeDelete(app, self, "nodeName_delete", module)

        self.enable_csv_export()

    def create_table(self, app, name, cls):
        return NameNodeTable(app, name, cls)

    def render_title(self, session):
        return "HDFS Name nodes"

class NameNodeTable(ObjectQmfSelectorTable):
    def __init__(self, app, name, cls):
        super(NameNodeTable, self).__init__(app, name, cls)

        self.id_col = self.NameNodeColumn(app, "idcol", cls.Id, cls.Id, ".namenodeframe")
        self.id_col.width = "10%"
        self.ipc_col = HadoopIpcColumn(app, "ipc", cls.Ipc, "20%")
        self.http_col = HadoopHttpColumn(app, "http", cls.Http, cls.Http, None)
        self.state_col = ObjectTableColumn(app, "statecol", cls.State)
        self.state_col.width = "10%"
        self.uptime_col = UptimeColumn(app, "uptimecol", cls.Uptime)
        self.uptime_col.width = "10%"
        self.owner_col = ObjectTableColumn(app, "ownercol", cls.Owner)

        self.add_column(self.id_col)
        self.add_column(self.ipc_col)
        self.add_column(self.http_col)
        self.add_column(self.state_col)
        self.add_column(self.uptime_col)
        self.add_column(self.owner_col)

    class NameNodeColumn(ObjectLinkColumn):
        def render_cell_href(self, session, record):
            id = unescape_entity(record[self.id_field.index])
            frame = self.frame.children_by_name[self.frame.name + self.frame_path]
            return frame.get_href(session, id, frame.parent.host.get(session))

class NameNodeCreate(Task):
    def __init__(self, app, name, module):
        super(NameNodeCreate, self).__init__(app, name)
        cls = app.model.com_redhat_cumin_grid_hadoop.NameNode

        self.module = module

        self.form = NameNodeCreateForm(app, self.name, self, cls)
        self.invoc = None

    def callback(self, result):
        if result == False:
            self.invoc.status = self.invoc.FAILED
        self.invoc.end()

    def do_invoke(self, session, object, invoc, args):
        self.invoc = invoc
        (binfile, owner, hadoophost, description) = args
        self.app.remote.start_name_node(hadoophost, binfile, owner, description, invoc.make_callback())

    def get_title(self, session, x):
        return "Create a name node"

    def enter(self, session, obj):
        form_session = wooly.Session(self.app.form_page)

        if obj:
            self.form.id.set(form_session, obj._id)

        self.form.hadoophost.set(form_session, self.module.frame.children_by_name[self.module.frame.name + ".ghschedframe"].host.get(session))
        self.form.return_url.set(form_session, session.marshal())
        self.form.show(form_session)

        self.do_enter(session, obj, form_session)
        return form_session


class NameNodeCreateForm(HadoopNodeCreateForm):
    def __init__(self, app, name, task, cls):
        super(NameNodeCreateForm, self).__init__(app, name, task, cls)

        self.binfile = BinFileLoc(app, "binfile")
        self.add_field(self.binfile)

        self.hadoophost = Parameter(app, "hadoophost")
        self.add_parameter(self.hadoophost)

    def process_display(self, session):
        self.scheduler.validate(session)

    def process_submit(self, session):
        self.validate(session)

        url = self.return_url.get(session)
        self.page.redirect.set(session, url)

        if not self.errors.get(session):
            binfile = self.binfile.get(session)
            hadoophost = self.hadoophost.get(session)
            owner = session.client_session.username()
            description = ""
            self.task.invoke(session, None, (binfile, owner, hadoophost, description))
            self.task.exit_with_redirect(session, url)

    def render_title(self, session):
        return "Create a name node"

class NameNodeDelete(HadoopNodeDeleteTask):
    def __init__(self, app, selector, name, module):
        super(NameNodeDelete, self).__init__(app, selector)
        cls = app.model.com_redhat_cumin_grid_hadoop.NameNode

        self.module = module

        self.form = NameNodeDeleteForm(app, self.name, self, cls)
        self.invoc = None

    def do_invoke(self, invoc, ids, args):
        self.invoc = invoc
        (hadoophost) = args
        self.app.remote.stop_name_node(hadoophost, ids, invoc.make_callback())

    def get_title(self, session):
        return "Remove name nodes"

    def get_item_content(self, session, item):
        return xml_escape(item)


class NameNodeDeleteForm(HadoopNodeDeleteForm):
    def process_submit(self, session):
        self.validate(session)
        ids = self.selection.get(session)
        hadoophost = self.hadoophost.get(session)
        url = self.return_url.get(session)
        self.page.redirect.set(session, url)

        if not self.errors.get(session):
            self.task.invoke(session, ids, (hadoophost))

    def render_title(self, session):
        return "Remove name nodes"

class NameNodeFrame(HadoopObjectFrame):
    def __init__(self, app, name, module):
        super(NameNodeFrame, self).__init__(app, name, module)

        self.view = NameNodeObjectView(app, "view", self.object)
        self.replace_child(self.view)

        self.data_node_frame = DataNodeFrame(app, "datanodeframe", module)
        self.add_mode(self.data_node_frame)

        self.overview = NameNodeOverview(app, self.name + ".nnoverview", self.object, self.module)
        self.view.add_tab(self.overview)

    def get_title(self, session):
        retval = "Name node:'%s' (%s)" % (xml_escape(self.object.get(session).Name), xml_escape(self.object.get(session).Id))
        return retval

    def get_object(self, session, id):
        host = self.frame.host.get(session)
        response = self.app.remote.get_name_node(self.frame.host.get(session), [id])
        obj = HadoopObject(response, self.app.model.com_redhat_cumin_grid_hadoop.NameNode)
        return obj

class NameNodeObjectView(HadoopObjectView):
    def add_details_tab(self):
        self.add_tab(NameNodeDetails(self.app, "details", self.object))

class NameNodeOverview(Widget):
    def __init__(self, app, name, node, module):
        super(NameNodeOverview, self).__init__(app, name)
        self.add_child(DataNodeSelector(app, self.name + ".dnfornn", module, False))

    def render_title(self, session):
        return "Data nodes"

class NameNodeDetails(HadoopObjectDetails):
    def __init__(self, app, name, object):
        super(NameNodeDetails, self).__init__(app, name, object)
        self.object = object

    def render_title(self, session):
        return "Details"

    def render_name(self, session):
        '''sample....would only be called if NameNodeDetails display was overridden to say {name}'''
        obj = self.object.get(session)
        return obj.Name
