from cumin.objectselector import ObjectSelector, ObjectLinkColumn, ObjectTable, \
    ObjectTableColumn, ObjectQmfTable, ObjectQmfSelectorTable, ObjectSelectorTaskForm, ObjectSelectorTask
from cumin.objectframe import ObjectFrame, ObjectTaskForm
from cumin.util import *
from cumin.qmfadapter import ObjectQmfAdapter
from cumin.task import Task, TaskLink
from cumin.objectframe import ObjectDetails

from wooly.forms import StringField

from hadoop import *
from hadoopobjectframe import *

from sage.util import *

from wooly.util import StringCatalog

strings = StringCatalog(__file__)

class TaskTrackerAdapter(HadoopAdapter):
    def get_sage_results(self, values):
        owner_filter = self.get_owner_filter(values['session'])
        results = self.app.remote.get_task_tracker_list(owner=owner_filter)
        return results

class TaskTrackerSelector(ObjectSelector):
    def __init__(self, app, name, module, check_owner=False):
        cls = app.model.com_redhat_cumin_grid_hadoop.TaskTracker

        super(TaskTrackerSelector, self).__init__(app, name, cls)

        self.module = module

        self.add_search_filter(self.table.id_col)
        self.table.adapter = TaskTrackerAdapter(app, cls, check_owner)

        task = TaskTrackerCreate(app, name + ".create", self.module)
        link = TaskLink(app, "taskTracker_create", task)
        self.links.add_child(link)

        TaskTrackerDelete(app, self, "taskTracker_delete", self.module)

        self.enable_csv_export()

    def create_table(self, app, name, cls):
        return TaskTrackerTable(app, name, cls)

    def render_title(self, session):
        return "Hadoop task trackers"

class TaskTrackerTable(ObjectQmfSelectorTable):
    def __init__(self, app, name, cls):
        super(TaskTrackerTable, self).__init__(app, name, cls)

        self.id_col = self.TaskTrackerColumn(app, "idcol", cls.Id, cls.Id, "tasktrackerframe")
        self.state_col = ObjectTableColumn(app, "statecol", cls.State)
        self.uptime_col = UptimeColumn(app, "uptimecol", cls.Uptime)
        self.owner_col = ObjectTableColumn(app, "ownercol", cls.Owner)

        # we need to be able to filter by parent_id, but don't need to see it
        self.parent_id_col = ObjectTableColumn(app, "parentid", cls.Parent_id)
        self.parent_id_col.visible = False

        self.add_column(self.id_col)
        self.add_column(self.state_col)
        self.add_column(self.uptime_col)
        self.add_column(self.owner_col)
        self.add_column(self.parent_id_col)

    def get_data_options(self, session):
        options = super(TaskTrackerTable, self).get_data_options(session)
        options.parent_filter = {"column":self.parent_id_col, "value":self.frame.id.get(session)}
        return options

    class TaskTrackerColumn(ObjectLinkColumn):
        def render_cell_href(self, session, record):
            id = unescape_entity(record[self.id_field.index])
            frame = self.frame.children_by_name[self.frame_path]
            return frame.get_href(session, id, frame.parent.host.get(session))

class TaskTrackerDelete(HadoopNodeDeleteTask):
    def __init__(self, app, selector, name, module):
        super(TaskTrackerDelete, self).__init__(app, selector)
        cls = app.model.com_redhat_cumin_grid_hadoop.TaskTracker

        self.module = module

        self.form = TaskTrackerDeleteForm(app, self.name, self, cls)
        self.invoc = None

    def do_invoke(self, invoc, ids, args):
        self.invoc = invoc
        (hadoophost) = args
        self.app.remote.stop_task_tracker(hadoophost, ids, invoc.make_callback())

    def get_title(self, session):
        return "Remove task trackers"

    def get_item_content(self, session, item):
        return xml_escape(item)


class TaskTrackerDeleteForm(HadoopNodeDeleteForm):
    def process_submit(self, session):
        self.validate(session)
        ids = self.selection.get(session)
        hadoophost = self.hadoophost.get(session)
        url = self.return_url.get(session)
        self.page.redirect.set(session, url)

        if not self.errors.get(session):
            self.task.invoke(session, ids, (hadoophost))

    def render_title(self, session):
        return "Remove task trackers"

class TaskTrackerCreate(Task):
    def __init__(self, app, name, module):
        super(TaskTrackerCreate, self).__init__(app, name)
        cls = app.model.com_redhat_cumin_grid_hadoop.TaskTracker

        self.module = module

        self.form = TaskTrackerCreateForm(app, self.name, self, cls)
        self.invoc = None

    def callback(self, result):
        if result == False:
            self.invoc.status = self.invoc.FAILED
        self.invoc.end()

    def do_invoke(self, session, object, invoc, args):
        self.invoc = invoc
        (binfile, owner, hadoophost, description, count, job_tracker) = args

        try:
            self.app.remote.start_task_tracker(hadoophost, job_tracker, binfile, owner, description, count, invoc.make_callback())
        except Exception, e:
            invoc.status = invoc.FAILED
            log.debug("Creating task trackers failed", exc_info=True)
            invoc.end()

    def enter(self, session, obj):
        form_session = wooly.Session(self.app.form_page)

        if obj:
            self.form.id.set(form_session, obj._id)

        self.form.hadoophost.set(form_session, self.app.gridhadoop.frame.children_by_name["gridhadoop.ghschedframe"].host.get(session))
        self.form.jt_id.set(form_session, self.app.gridhadoop.frame.children_by_name["gridhadoop.ghschedframe"].children_by_name["gridhadoop.ghschedframe.jobtrackerframe"].id.get(session))
        self.form.return_url.set(form_session, session.marshal())
        self.form.show(form_session)

        self.do_enter(session, obj, form_session)
        return form_session

    def get_title(self, session, x):
        return "Create task trackers"


class TaskTrackerCreateForm(HadoopNodeCreateForm):
    def __init__(self, app, name, task, cls):
        super(TaskTrackerCreateForm, self).__init__(app, name, task, cls)

        self.binfile = BinFileLoc(app, "binfile")
        self.add_field(self.binfile)

        self.count = self.CountField(app, "count")
        self.add_field(self.count)

        self.hadoophost = Parameter(app, "hadoophost")
        self.add_parameter(self.hadoophost)

        self.jt_id = Parameter(app, "jt_id")
        self.add_parameter(self.jt_id)

    def process_display(self, session):
        self.scheduler.validate(session)

    def process_submit(self, session):
        self.validate(session)

        url = self.return_url.get(session)
        self.page.redirect.set(session, url)

        if not self.errors.get(session):
            binfile = self.binfile.get(session)
            owner = session.client_session.username()
            hadoophost = self.hadoophost.get(session)
            count = self.count.get(session)
            job_tracker = self.jt_id.get(session)
            description = ""
            self.task.invoke(session, None, (binfile, owner, hadoophost, description, count, job_tracker))
            self.task.exit_with_redirect(session, url)

    def render_title(self, session):
        return "Create task trackers"

    class CountField(IntegerField):
        def __init__(self, app, name):
            super(TaskTrackerCreateForm.CountField, self).__init__(app, name)
            self.input.param.default = 1
            self.input.size = 4

        def render_title(self, session):
            return "Number of task trackers to start"


class JobTrackerField(ScalarField):
    def __init__(self, app, name):
        super(JobTrackerField, self).__init__(app, name, None)

        self.param = IntegerParameter(app, "param")
        self.add_parameter(self.param)

        cls = app.model.com_redhat_cumin_grid_hadoop.JobTracker

        self.input = self.JobTrackerOptions(app, "input", self.param)
        self.add_child(self.input)

    def get(self, session):
        return self.input.get(session)

    def validate(self, session):
        super(JobTrackerField, self).validate(session)

        job_tracker = self.input.get_items(session)

        if not job_tracker:
            error = FormError("There is no job tracker to bind to")
            self.form.errors.add(session, error)

    def render_title(self, session):
        return "Job tracker"

    def render_help(self, session):
        return "Attach job trackers to this job tracker"

    class JobTrackerOptions(OptionInputSet):
        def do_process(self, session):
            cls = self.app.model.com_redhat_cumin_grid_hadoop.JobTracker
            id = self.param.get(session)

            if id is None:
                items = self.get_items(session)

                if items:
                    self.param.set(session, items[0])

            super(JobTrackerField.JobTrackerOptions, self).do_process \
                (session)

        def do_get_items(self, session):
            nodes = []
            (status, node_list) = self.app.remote.get_job_tracker_list()
            for node in node_list:
                nodes.append(node.ref)
            return sorted(nodes, cmp=compareHadoopIds)

        def render_item_value(self, session, item):
            value = None
            if item.ipc:
                value = item.ipc
            else:
                value = item.id
            return value

        def render_item_content(self, session, item):
            display_string = "%s (%s)" % (item.ipc, item.id)
            return xml_escape(display_string)

        def render_item_selected_attr(self, session, item):
            if item == self.param.get(session):
                return "selected=\"selected\""


class TaskTrackerFrame(HadoopObjectFrame):
    def __init__(self, app, name, module):
        super(TaskTrackerFrame, self).__init__(app, name, module)

        self.module = module

        self.view = TaskTrackerObjectView(app, "view", self.object)
        self.replace_child(self.view)

        self.details = TaskTrackerDetails(self.app, self.name + ".details", self.object)
        self.view.add_tab(self.details)

    def get_title(self, session):
        retval = ""
        try:
            retval = "Task tracker:'%s' (%s)" % (xml_escape(self.object.get(session).Name), xml_escape(self.object.get(session).Id))
        except Exception, e:
            pass
        return retval

    def get_object(self, session, id):
        host = self.frame.host.get(session)
        response = self.app.remote.get_task_tracker(self.frame.host.get(session), [id])
        obj = HadoopObject(response, self.app.model.com_redhat_cumin_grid_hadoop.TaskTracker)
        return obj

class TaskTrackerObjectView(HadoopObjectView):
    def add_details_tab(self):
        pass

    def render_title(self, session):
        obj = self.object.get(session)
        if obj is not None:
            return "Task tracker: %s" % xml_escape(obj.Id)
        else:
            return ""

class TaskTrackerDetails(HadoopObjectDetails):
    def __init__(self, app, name, object):
        super(TaskTrackerDetails, self).__init__(app, name, object)
        self.object = object

    def render_title(self, session):
        return "Details"

    def render_name(self, session):
        '''sample....would only be called if TaskTrackerDetails display was overridden to say {name}'''
        obj = self.object.get(session)
        return obj.Name
