import os
import threading
import logging
import random
import urllib2
import string
import time
import sage
import socket

from datetime import datetime
from threading import Lock
from suds import *
from sage.util import MethodResult, CallSync, CallThread, host_list,\
    parse_URL, get_datadir
from sage.catalog import Catalog
from aviarylocator import AviaryLocator
from clients import ClientPool, TransportFactory

log = logging.getLogger("sage.aviary")

class SudsLogging(object):
    '''
    Class to enable logging of SOAP transactions.

    This class allows detailed logging of SOAP transactions through
    the Suds module.  It is meant as a debugging aid during development.
    '''
    _on = False
    sudslogs = {"suds.client": None, 
                "suds.transport": None, 
                "suds.xsd.schema": None, 
                "suds.wsdl": None}

    @classmethod
    def set(cls, flag, home):
        if flag:
            SudsLogging._on = True
            for k,v in SudsLogging.sudslogs.iteritems():
                l = logging.getLogger(k)
                l.setLevel(logging.DEBUG)
                try:
                    f = open(os.path.join(home, "log/"+k+".log"), 'a+')
                    h = logging.StreamHandler(f)
                    l.addHandler(h)
                    SudsLogging.sudslogs[k] = h
                except:
                    pass
            log.debug("AviaryOperations: suds logging on")

        elif SudsLogging._on:
            SudsLogging._on = False
            for k,v in SudsLogging.sudslogs.iteritems():
                try:
                    logging.getLogger(k).removeHandler(v)
                except:
                    pass
            log.debug("AviaryOperations: suds logging off")


#Stuff to do:

#- Test pool/schedd/stuff with hierarchical collectors

#- Add a summary comment at the top like QMF opreations

#- can we use default/timeout with suds?
#looks like the way to do this is through the Transport object that
#is set in the client.  The timeout is set on the Transport itself,
#and the Transport can be changed on the client with set_options.
#Alternatively, we could retain a reference to the Transport in 
#the clients that we pool so that we can call set_options on the
#transport.

class BaseServerList(object):
    # Nice, friendly strings for error messages on lookup
    _nice = {"JOB": "job service",
             "QUERY_SERVER": "query service",
             "HADOOP": "hadoop service"}

    def __init__(self, subtype):
        self.servers = None
        self.should_retry = False

        try:
            self.nice = BaseServerList._nice[subtype]
        except:
            self.nice = subtype

    def _get_host_set(self):
        hosts = set()
        for k,v in self.servers.iteritems():
            for url in v:
                hosts.add(k+":"+url.port)
        return hosts

    def _get_host(self, name):
        '''
        Lookup a host in a dictionary produced by sage.util.host_list.
        The host may be just a hostname or hostname:port.
        Return the scheme and URL for the host.  The scheme is separated
        out for convenience (although it could be parsed from the URL)
        '''
        scheme = ""
        host = ""
        port = None

        # Support hostnames with ports, too.
        name = socket.getfqdn(name)
        if ":" in name:
            u = parse_URL(name)
            name = u.host
            port = u.port

        if name in self.servers:
            url = None
            urls = self.servers[name]
            if port is not None:
                # find the matching port
                for u in urls:
                    if u.port == port:
                        url = u
                        break

            elif len(urls) > 0:
                url = random.sample(urls, 1)[0]

            if url: 
                scheme = url.scheme
                host = str(url)
                # A particular method name is going to be appended to path,
                # so ensure the final "/" here.
                if not host.endswith("/"):
                    host += "/"
        return scheme, host
            
    def get_hosts(self):
        return self._get_host_set()

class ServerList(BaseServerList):
    '''
    Query an Aviary locator object for endpoints by reource and subtype.
    Lookup an endpoint of the specified type by hostname or hostname:port.
    Allow the cached list to be refreshed on demand.
    '''
    def __init__(self, locator, resource, subtype):
        super(ServerList, self).__init__(subtype)

        # Since we have a dynamic server list, failed operations
        # may be retried
        self.should_retry = True

        self._lock = Lock()
        self.locator = locator
        self.resource = resource
        self.subtype = subtype

    def _refresh(self):
        log.debug("AviaryOperations: refresh server list for %s %s" \
                      % (self.resource, self.subtype))
        try:
            result = self.locator.get_endpoints(self.resource, 
                                                self.subtype)
        except Exception, e:
            result = _AviaryCommon._pretty_result(e, 
                                             self.locator.locator_uri)
            log.debug("AviaryOperations: failed to get endpoints, " \
                      "exception message '%s'" % result.message)
            raise result

        urls = []
        # If it's not okay, we just produce an empty sever list
        if result.status.code == "OK":
            for r in result.resources:
                urls.extend(r.location)
        urls = ",".join(urls)
        if urls == "":
            log.info("AviaryOperations: locator returned " \
                     "no endpoints for %s %s" % (self.resource,
                                                 self.subtype))

        # Parse the urls and produce a dictionary of
        # sage_URLs by machine
        self.servers = host_list(urls)

    def _find_server(self, machine, refresh=False):
        '''
        Search the cached server list for machine using _get_host.
        If the server list is empty or refresh is True, get a new list
        of endpoints from the Aviary locator object and generate a new
        server list.
        '''
        # If we already have a list of values then return that list unless
        # refresh is True
        self._lock.acquire()
        try:
            if self.servers is None or refresh:
                self._refresh()
            scheme, host = self._get_host(machine)
        finally:
            self._lock.release()
        return scheme, host

    def find_server(self, machine, refresh=False):
        '''
        Lookup a URL in the cached server list by hostname or hostname:port.
        Generate a new server list if refresh is True, or if the
        initial host lookup fails and refresh == "on_no_host".
        Return the scheme and URL for the specified host if found or raise
        an exception if the host could not be found.
        '''
        # Update the list if necessary and return host info for machine.
        # If refresh is True, the server list will be updated.  This is
        # used when a previously cached host fails during a remote method
        # invocation and we want to do a retry before giving up.
        scheme, host = self._find_server(machine, refresh is True)
        if host == "" and refresh == "on_no_host":
            # Okay, if we didn't find a host and on_no_host has been set,
            # try one more time with a forced refresh before we error.
            # The case here is that the service on the particular host has
            # not yet been discovered since Cumin startup or the last refresh.
            # But, the service may appear at any point so it's possible that
            # it's there now.  Think "on demand polling".
            scheme, host = self._find_server(machine, True)
        if host == "":
            log.info("AviaryOperations: failed to locate %s on %s" \
                         % (self.nice, machine))
            raise Exception("Cannot locate %s on %s via aviary locator" \
                                % (self.nice, machine))
        return scheme, host

    def get_hosts(self):
        self._lock.acquire()
        try:
            self._refresh()
            res = self._get_host_set()
        finally:
            self._lock.release()
        return res

class FixedServerList(BaseServerList):
    '''
    Query a fixed server list for Aviary endpoints.
    Lookup an endpoint in the fixed list by hostname or hostname:port.
    The fixed server list is an alternative to using an Aviary locator.
    '''
    def __init__(self, servers, port, path, subtype):
        '''
        Initialize a fixed server list of Aviary endpoints.
        servers -- a comma separated list of (partial) URLs
        port -- default port value if a URL does not specify a port
        path -- default path value if a URL does not specify a path
        subtype -- this is used to identify the service type in error
           messages.  Predefind values are "JOB" and "QUERY_SERVER".
        The scheme for a URL will be "http" if not specified.
        '''
        super(FixedServerList, self).__init__(subtype)

        # Replace any occurrence of locahost with output of gethostname()
        # before parsing to match Machine fields of QMF objects later on.
        host = socket.gethostname()
        servers = string.replace(servers, "localhost", host)

        self.servers = host_list(servers, 
                                 default_scheme="http", 
                                 default_port=port,
                                 default_path=path)

    def find_server(self, machine, *args):
        '''
        Lookup a URL in the fixed server list by hostname or hostname:port.
        Return the scheme and URL for the specified host if found or raise
        an exception if the host could not be found.
        '''
        scheme, host = self._get_host(machine)
        if host == "":
            log.info("AviaryOperations: failed to locate %s on %s" \
                         % (self.nice, machine))
            raise Exception("Cannot locate %s on %s, check aviary " \
                            "settings in cumin.conf" % (self.nice, machine))
        return scheme, host

class MockScheduler(object):
    '''
    Pseudo scheduler object.

    Most methods in the _AviaryXXX objects were written to be compatible
    with QMF methods which take Rosemary objects.

    This is a mock scheduler object containing the fields 
    referenced by the methods: a scheduler name, a pool identifier, 
    and a machine hostname/ip.

    This object may be used when a scheduler object from Rosemary
    is not available (ie, sage is being used outside of Cumin).  The
    only required value is machine
    '''
    def __init__(self, machine, name = None, pool = None):
        self.Name = name
        self.Pool = pool
        self.Machine = machine


class MockSubmission(object):
    '''
    Pseudo submission object.

    Most methods in the _AviaryXXX objects were written to be compatible
    with QMF methods which take a Rosemary object.

    This is a mock submission object containing the fields 
    referenced by the methods: a submission name and a submission owner.

    This object may be used when a submission object from Rosemary
    is not available (ie, sage is being used outside of Cumin).  The
    only required value is name.
    '''
    def __init__(self, name, owner = None):
        self.Name = name
        self.Owner = owner

class _AviaryCommon(object):
    def __init__(self, transports, servers, wsdl, resource, subtype):

        self.transports = transports
        self.servers = servers
        self.resource = resource
        self.subtype = subtype
        self.client_pool = ClientPool(wsdl, None)

        # Things in Aviary that were born as swap-ins/analogs for
        # existing QMF methods need to pass synchronous results back
        # in a MethodResult object because that's what code expected
        # from QMF.  However, newer stuff (like hadoop) doesn't need
        # to wrap results this way.  Let it be controlled.
        self.use_MethodResult_for_sync_calls = True

    def get_hosts(self, resource, subtype):
        if resource == self.resource and \
           (not subtype or self.subtype == subtype):
            return self.servers.get_hosts()
        return set()    

    def _set_client_info(self, client, refresh=False):
        # See if the server_name is a full URL.  If so, we'll use it.
        url = parse_URL(client.server_name)
        if url.scheme and url.host and url.port and url.path:
            scheme = url.scheme
            host = client.server_name
        
        elif client.server_list:
            scheme, host = client.server_list.find_server(client.server_name, 
                                                          refresh)
        else:
            log.debug("AviaryOperations: no server list specified and %s "\
                     "is not a URL" % client.server_name) 
            raise Exception("Cannot resolve %s to an Aviary endpoint" % \
                                                      client.server_name)

        # Have to set the URL for the method.  This might go away someday...
        client.set_options(location=host+client.method_name)

        # Since we pool the clients and reuse them for different requests
        # and since its possible to be using servers with different schemes,
        # we have to always reset the transport here.
        the_transport = self.transports.get_transport(scheme)
        client.set_options(transport=the_transport)

    def _setup_client(self, client, server_list, name, meth_name):
        # Look up the host and construct the URL.
        # Store information in the client so that retry is possible.
        client.server_list = server_list
        client.server_name = name
        client.method_name = meth_name

        # This is initial setup before a call so we want to try a 
        # refresh on the server list if our service comes up missing
        self._set_client_info(client, refresh="on_no_host")
        return client

    @classmethod
    def _get_status(cls, result):
        # For Aviary operations, if the operation
        # did not work the reason is in the text field.
        # In cumin, we want to pass any error text as
        # the status parameter to callbacks
        if result.code != "OK":
            return result.text
        return result.code

    @classmethod
    def _cb_args_dataless(cls, result):
        # Marshal data in result for passing to standard callback.
        # This routine is for results that contain status only, no data.
        if isinstance(result, Exception):
            status = result
        else:
            status = _AviaryCommon._get_status(result)
        return (status, None)

    @classmethod
    def _pretty_result(cls, result, host):
        if isinstance(result, urllib2.URLError):
            return Exception("Trouble reaching host %s, %s" % (host, result.reason))
        elif isinstance(result, Exception):
            if hasattr(result, "args"):
                reason = result.args
            else:
                reason = str(result)
            return Exception("Operation failed on host %s, %s" % (host, reason))
        return result

    def call_client_retry(self, client, meth_name, *meth_args, **meth_kwargs):
        # If we fail with a urllib2.URLError (or anything similar) then
        # attempt to get a new endpoint and try again.
        meth = getattr(client.service, meth_name)
        try:
            result = meth(*meth_args, **meth_kwargs)
        except Exception, e:
            # If we get an exception, our endpoint may have moved 
            # (probably due to a restart on the condor side)
            # Let's get new endpoints, reset the client, 
            # and try again.
            if client.server_list and client.server_list.should_retry:
                log.debug("AviaryOperations: received %s, retrying %s"\
                              % (str(e), client.options.location))
                self._set_client_info(client, refresh=True)
                result = meth(*meth_args, **meth_kwargs)
            else:
                raise e
        return result

    def _call_sync(self, meth, process_results, *meth_args, **meth_kwargs):
        # Common interface with QMF operations requires that a MethodResult
        # object (or something just like it) be returned for synchronous calls.
        # Might as well use the CallSync since the completion() routine 
        # handles status semantics.

        # meth here is a suds.client.Method which has a name attr....
        sync = CallSync(log, name=meth.__name__)
        try:
            result = meth(*meth_args, **meth_kwargs)
        except Exception, e:
            result = e
        if self.use_MethodResult_for_sync_calls:
            sync.get_completion()(*process_results(result)) 
            return MethodResult(sync)
        else:
            return process_results(result)

class _AviaryJobMethods(_AviaryCommon):
    def __init__(self, locator, transports, datadir, job_servers):

        resource = "SCHEDULER"
        subtype = "JOB"
        if locator:
            servers = ServerList(locator, 
                                 resource, subtype)
        elif job_servers:
            servers = FixedServerList(job_servers, 
                                      "9090", 
                                      "/services/job/",
                                      subtype)
        else:
            servers = None        

        wsdl = "file:" + os.path.join(get_datadir(datadir, "job"),
                                      "aviary-job.wsdl")
        
        super(_AviaryJobMethods, self).__init__(transports,
                                                servers,
                                                wsdl,
                                                resource,
                                                subtype)        

    def set_job_attribute(self, scheduler, job_id, name, value, callback, submission):
        assert callable(callback)

        def my_callback(result):
            self.client_pool.return_object(job_client)
            result = self._pretty_result(result, scheduler.Machine)
            # massage results for use by standard callback
            callback(*self._cb_args_dataless(result))

        job_client = self.client_pool.get_object()
        self._setup_client(job_client, 
                           self.servers,   # server lookup object
                           scheduler.Machine,  # host we want
                           "setJobAttribute")
                                   
        # Make a job id parameter (see job wsdl)
        jobId = job_client.factory.create('ns0:JobID')
        jobId.job = job_id
        jobId.pool = scheduler.Pool
        jobId.scheduler = scheduler.Name
        jobId.submission.name = submission.Name
        jobId.submission.owner = submission.Owner

        # Make attribute parameter from name and value
        aviary_attr = job_client.factory.create('ns0:Attribute')
        aviary_attr.name = name
        aviary_attr.type = "STRING"
        aviary_attr.value = value

        t = CallThread(self.call_client_retry, my_callback, 
                       job_client, "setJobAttribute", jobId, aviary_attr)
        t.start()

    def submit_job(self, scheduler, ad, callback):
        assert callable(callback)

        type_to_aviary = {int: "INTEGER", 
                          float: "FLOAT", 
                          str: "STRING", 
                          bool: "BOOLEAN"}

        def my_callback(result):
            # Turn this back off before we put it back in the pool
            # so allow_overrides isn't set for someone else...
            job_client.set_enable_attributes(False)
            self.client_pool.return_object(job_client)
            data = None
            result = self._pretty_result(result, scheduler.Machine)
            if isinstance(result, Exception):
                status = result                
            else:
                # the aviary response has the job id available,
                # we'll pass it anyway even though Cumin does not care
                # at the present time
                status = _AviaryCommon._get_status(result.status)
                if status == "OK" and hasattr(result, "id"):
                    data = result.id
            callback(status, data)

        job_client = self.client_pool.get_object()
        self._setup_client(job_client, 
                           self.servers,    # server lookup object
                           scheduler.Machine,  # host we want
                           "submitJob")

        # Set basic attributes in the order defined by aviary-job.wsdl.
        args = list()
        basic_attrs = ("Cmd", "Args", "Owner", "Iwd", "Submission")
        for attr in basic_attrs:
            try:
                args.append(ad[attr])
            except:
                # Someone may be unhappy if this is a required param!
                # Let the downstream code generate an error
                args.append("")

        # Add empty list for Aviary's basic requirement value...
        args.append([])

        # and let's let Requirements remain an unrestricted expression so that
        # we can just pass through the value from Cumin without interfering.
        # To do that, we need to specify Requirements through the
        # "extras" fields and set allowOverrides to True.
        # (otherwise, Requirements will be limited to particular
        # resource constraint types defined by aviary)
        job_client.set_enable_attributes(True)
        job_client.set_attributes({"allowOverrides": "true"})
        extras = list()
        for k, v in ad.iteritems():
            # We don't need to send descriptors down to aviary
            # and basic_attrs have already been filled in
            if k == "!!descriptors" or k in basic_attrs:
                continue
            
            extra = job_client.factory.create('ns0:Attribute')
            extra.name = k
            # But we do need to look in descriptors to find expressions...
            if k in ad["!!descriptors"]:
                extra.type = "EXPRESSION"
            else:
                try:
                    extra.type = type_to_aviary[type(v)]
                except KeyError:
                    extra.type = "UNDEFINED"
            extra.value = v
            extras.append(extra)

        # Important, extras itself must be added as an embedded list or 
        # suds will consider only a single item
        args.append(extras)

        t = CallThread(self.call_client_retry, my_callback, 
                       job_client, "submitJob", *args)
        t.start()

    def control_job(self, cmd, scheduler, job_id, reason, submission, *args, **kwargs):
        '''
        This method is asynchronous iff 'callback' is supplied.

        Values for cmd are case sensitive (although the first letter may
        actually be either case) and may be one of the following:
            'holdJob', 'releaseJob', 'suspendJob', 'continueJob', 'removeJob'
        
        kwargs will be searched for 'callback', 'default' and 'timeout' arguments.
        '''

        # Aviary and QMF command names for job control differ in one respect,
        # which is that QMF uses an initial capital and Aviary does not.
        cmd = cmd[0].lower() + cmd[1:]
        return self._control_job(scheduler, job_id, reason, submission,
                                 cmd, *args, **kwargs)

    def _control_job(self, scheduler, job_id, reason, submission,
                     meth_name, *args, **kwargs):

        callback = "callback" in kwargs and kwargs["callback"] or None
        default = "default" in kwargs and kwargs["default"] or None
        timeout = "timeout" in kwargs and kwargs["timeout"] or 5

        client = self.client_pool.get_object()
        self._setup_client(client, 
                           self.servers,    # server lookup object
                           scheduler.Machine,  # host we want
                           meth_name)

        # Make a job id parameter (see job wsdl)
        jobId = client.factory.create('ns0:JobID')
        jobId.job = job_id
        jobId.pool = scheduler.Pool
        jobId.scheduler = scheduler.Name
        jobId.submission.name = submission.Name
        jobId.submission.owner = submission.Owner

        if callback:
            def my_callback(result):
                self.client_pool.return_object(client)
                # Fix up the exception message if necessary
                result = self._pretty_result(result, scheduler.Machine)
                callback(*self._cb_args_dataless(result))

            t = CallThread(self.call_client_retry, my_callback,
                           client, meth_name, jobId, reason)
            t.start()
        else:
            def my_process_results(result):
                # Fix up the exception message if necessary
                result = self._pretty_result(result, scheduler.Machine)
                return self._cb_args_dataless(result)

            res = self._call_sync(self.call_client_retry, my_process_results, 
                                  client, meth_name, jobId, reason) 
            self.client_pool.return_object(client)
            return res;

class _AviaryQueryMethods(_AviaryCommon):
    def __init__(self, locator, transports, datadir, query_servers):

        resource = "CUSTOM"
        subtype = "QUERY_SERVER"
        if locator:
            servers = ServerList(locator, 
                                 resource, subtype)
        elif query_servers:
            servers = FixedServerList(query_servers, 
                                      "9091", 
                                      "/services/query/", 
                                      subtype)
        else:
            servers = None

        wsdl = "file:" + os.path.join(get_datadir(datadir,"query"), 
                                      "aviary-query.wsdl")

        super(_AviaryQueryMethods, self).__init__(transports,
                                                  servers,
                                                  wsdl,
                                                  resource,
                                                  subtype)

    def fetch_job_data(self, job_server, job_id, ftype, file, start, end,
                       scheduler_name, submission, *args, **kwargs):
        '''
        kwargs will be searched for 'default' and 'timeout' arguments.

        These optional arguments were moved to kwargs for compatibility
        with another implementation of the same routine.
        '''

        default = "default" in kwargs and kwargs["default"] or None
        timeout = "timeout" in kwargs and kwargs["timeout"] or 5

        # Aviary doesn't use the file name as does QMF, instead it
        # specifies the file type and lets condor figure out the path.

        def my_process_results(result):
            # Fix up the exception message if necessary
            data = None
            result = self._pretty_result(result, job_server.Machine)
            if isinstance(result, Exception):
                status = result
            else:
                status = _AviaryCommon._get_status(result.status)
                if status == "OK" and hasattr(result, "content"):
                    # Match the format expected by Cumin.  This is
                    # the format used by the QMF call...
                    data = {'Data': result.content}
            return (status, data)

        client = self.client_pool.get_object()
        self._setup_client(client, 
                           self.servers,  # server lookup object
                           job_server.Machine,  # host we want
                           "getJobData")

        # Make a job data parameter (see query wsdl)
        jobData = client.factory.create('ns0:JobData')
        jobData.id.job = job_id
        jobData.id.pool = job_server.Pool
        jobData.id.scheduler = scheduler_name
        jobData.id.submission.name = submission.Name
        jobData.id.submission.owner = submission.Owner

        # Translate cumin file type to Aviary file type
        if ftype == "e":
            jobData.type = "ERR"
        elif ftype == "o":
            jobData.type = "OUT"
        elif ftype == "u":
            jobData.type = "LOG"
        else:
            # We can't translate the type.
            # Let Aviary throw an error on this instead of us.
            jobData.type = ftype

        from_end = start < 0
        max_bytes = abs(end - start)

        res = self._call_sync(self.call_client_retry, my_process_results,
                              client, "getJobData", jobData, max_bytes, from_end)
        self.client_pool.return_object(client)
        return res;

    def get_job_ad(self, job_server, job_id, scheduler_name, submission, 
                   *args, **kwargs):
        '''
        kwargs will be searched for 'default' and 'timeout' arguments.

        These optional arguments were moved to kwargs for compatibility
        with another implementation of the same routine.
        '''        

        default = "default" in kwargs and kwargs["default"] or None
        timeout = "timeout" in kwargs and kwargs["timeout"] or 5

        aviary_to_type = {"INTEGER": int, 
                          "FLOAT": float, 
                          "STRING": str, 
                          "BOOLEAN": bool}

        def make_tuple(attr):
            # Attempt to cast the value into the specified type
            try:
                v = aviary_to_type[attr.type](attr.value)
            except:
                v = attr.value
            return (attr.name, v)
        
        def make_dict(attrs):
            return dict([make_tuple(attr) for attr in attrs])

        def my_process_results(result):
            data = None
            # Fix up the exception message if necessary
            result = self._pretty_result(result, job_server.Machine)
            if isinstance(result, Exception):
                status = result
            else:
                status = _AviaryCommon._get_status(result[0].status)
                if status == "OK":
                    # Match the format expected by Cumin.  This is
                    # the format used by the QMF call.  We have a list
                    # of attributes in attrs that we need to make into
                    # a dictionary
                    ads = make_dict(result[0].details.attrs) 
                    data = {'JobAd': ads}
            return (status, data)

        client = self.client_pool.get_object()
        self._setup_client(client, 
                           self.servers,  # server lookup object
                           job_server.Machine,  # host we want
                           "getJobDetails")

        # Make a job id parameter (see job wsdl)
        jobId = client.factory.create('ns0:JobID')
        jobId.job = job_id
        jobId.pool = job_server.Pool
        jobId.scheduler = scheduler_name
        jobId.submission.name = submission.Name
        jobId.submission.owner = submission.Owner

        res = self._call_sync(self.call_client_retry, my_process_results,
                              client, "getJobDetails", jobId)
        self.client_pool.return_object(client)
        return res;

    def get_job_summaries(self, submission, callback, machine_name):
        assert callable(callback)

        def to_int_seconds(dt):
            # Change a datetime.datetime into int seconds since epoch
            # Note, this works nicely if the datetime happens to include microseconds
            # since the call to timetuple will drop them.  Stuff coming back from
            # condor should not have microseconds anyway.
            return int(time.mktime(dt.timetuple()))

        def get_string(job, attr):
            # Cast suds text types into str so we have standard Py types
            # Handles optional strings as well
            if hasattr(job, attr):
                return str(getattr(job, attr))
            return ""

        def adapt(jobs):
            # Make an aviary job summary look like the canonical form
            # that cumin is expecting (actually the QMF form because of history).
            result = list()
            for job in jobs:
                cluster, proc = job.id.job.split(".")
                j = dict()
                j["ClusterId"]            = int(cluster)
                j["Cmd"]                  = str(job.cmd)
                j["EnteredCurrentStatus"] = to_int_seconds(job.last_update)

                # Note, GlobalJobId here will not match the same value from
                # QMF because the qdate portion of the name is missing
                j["GlobalJobId"]          = job.id.scheduler + \
                                            "#" + job.id.job

                j["JobStatus"]            = str(job.job_status)
                j["ProcId"]               = int(proc)
                j["QDate"]                = to_int_seconds(job.queued)
                
                # These may be null...
                j["Args"]                 = get_string(job, "args1")
                j["ReleaseReason"]        = get_string(job, "released")
                j["HoldReason"]           = get_string(job, "held")
                result.append(j)
            return result
                
        def my_callback(result):
            query_client.set_enable_attributes(False)
            self.client_pool.return_object(query_client)
            result = self._pretty_result(result, machine_name)
            if isinstance(result, Exception):
                callback(result, None)                
            else:
                status =  _AviaryCommon._get_status(result[0].status)
                if status == "OK" and hasattr(result[0], "jobs"):
                    data = {"Jobs": adapt(result[0].jobs)}
                else:
                    data = {"Jobs": None}
                callback(status, data)

        query_client = self.client_pool.get_object()
        self._setup_client(query_client, 
                           self.servers,  # server lookup object
                           machine_name,        # host we want
                           "getSubmissionSummary")

        # What we really want here is the job summaries from the
        # submission summary response.  To get those, we have to
        # set an extra attribute on the client...
        query_client.set_enable_attributes(True)
        query_client.set_attributes({"includeJobSummaries": "true"})

        # Make a submission id.  (see query wsdl)
        subId = query_client.factory.create('ns0:SubmissionID')
        subId.name = submission.Name
        subId.owner = submission.Owner

        t = CallThread(self.call_client_retry, my_callback, 
                       query_client, "getSubmissionSummary", subId)
        t.start()

class _AviaryHadoopMethods(_AviaryCommon):
    def __init__(self, locator, transports, datadir, hadoop_servers):

        resource = "SCHEDULER"
        subtype = "HADOOP"
        if locator:
            servers = ServerList(locator, 
                                 resource, subtype)
        else:
            servers = FixedServerList(hadoop_servers, 
                                      "9090", 
                                      "/services/hadoop/",
                                      subtype)        

        wsdl = "file:" + os.path.join(get_datadir(datadir, "hadoop"),
                                      "aviary-hadoop.wsdl")

        super(_AviaryHadoopMethods, self).__init__(transports,
                                                   servers,
                                                   wsdl,
                                                   resource,
                                                   subtype)

        # Equivalence? We don't want no stinking QMF structural equivalence!
        self.use_MethodResult_for_sync_calls = False

    def start_name_node(self, host, bin_file, owner, description, callback):
        assert callable(callback)

        client = self.client_pool.get_object()
        self._setup_client(client,
                           self.servers,
                           host,
                           "startNameNode")

        def result_tuple(result, host):
            data = None
            result = self._pretty_result(result, host)
            if isinstance(result, Exception):
                status = result
            else:
                status = _AviaryCommon._get_status(result.status)
                if status == "OK" and hasattr(result, "ref"):
                    data = result.ref
            return (status, data)

        def my_callback(result):
            self.client_pool.return_object(client)
            callback(*result_tuple(result, host))

        t = CallThread(self.call_client_retry, my_callback,
                       client, "startNameNode", bin_file, owner, description)
        t.start()

    def stop_name_node(self, host, ids, callback):
        assert callback
        self._operate_on_ids(host, ids, callback, "stopNameNode")

    def get_name_node(self, host, ids, owner=None, callback=None):
        return self._query_ids(host, ids, owner, callback, "getNameNode")

    def start_data_node(self, host, nn_id, bin_file, owner, description, count, callback):
        return self._start_node(host, nn_id, bin_file, owner, description, count,
                                "startDataNode", callback)

    def stop_data_node(self, host, ids, callback):
        self._operate_on_ids(host, ids, callback, "stopDataNode")

    def get_data_node(self, host, ids, owner=None, callback=None):
        return self._query_ids(host, ids, owner, callback, "getDataNode")

    def start_job_tracker(self, host, nn_id, bin_file, owner, description, count, callback):
        return self._start_node(host, nn_id, bin_file, owner, description, count,
                                "startJobTracker", callback)

    def stop_job_tracker(self, host, ids, callback):
        self._operate_on_ids(host, ids, callback, "stopJobTracker")

    def get_job_tracker(self, host, ids, owner=None, callback=None):
        return self._query_ids(host, ids, owner, callback, "getJobTracker")

    def stop_task_tracker(self, host, ids, callback):
        self._operate_on_ids(host, ids, callback, "stopTaskTracker")

    def start_task_tracker(self, host, nn_id, bin_file, owner, description, count, callback):
        return self._start_node(host, nn_id, bin_file, owner, description, count,
                                "startTaskTracker", callback)

    def get_task_tracker(self, host, ids, owner=None, callback=None):
        return self._query_ids(host, ids, owner, callback, "getTaskTracker")

    def get_name_node_list(self, owner=None, callback=None):
        return self._get_node_list(self.get_name_node, owner, callback)

    def get_data_node_list(self, owner=None, callback=None):
        return self._get_node_list(self.get_data_node, owner, callback)

    def get_job_tracker_list(self, owner=None, callback=None):
        return self._get_node_list(self.get_job_tracker, owner, callback)

    def get_task_tracker_list(self, owner=None, callback=None):
        return self._get_node_list(self.get_task_tracker, owner, callback)    

    def _make_id(self, client, val):

        def url(v):
            try:
                return "://" in v
            except Exception:
                pass
            return False

        ref = client.factory.create("ns1:HadoopID")
        try:
            if url(val):
                ref.ipc = val
            else:
                ref.id = str(val)
        except Exception:
            raise Exception("Invalid HadoopID value")       
        return ref

    def _make_id_list(self, client, vals):
        refs = []
        for v in vals:
            ref = self._make_id(client, v)
            refs.append(ref)        
        return refs

    def _start_node(self, host, ref_id, bin_file, owner, description, count, 
                    meth_name, callback):
        assert callable(callback)

        client = self.client_pool.get_object()
        self._setup_client(client,
                           self.servers,
                           host,
                           meth_name)

        def result_tuple(result, host):
            data = None
            result = self._pretty_result(result, host)
            if isinstance(result, Exception):
                status = result
            else:
                status = _AviaryCommon._get_status(result.status)
                if status == "OK" and hasattr(result, "ref"):
                    data = result.ref
            return (status, data)

        def my_callback(result):
            self.client_pool.return_object(client)
            callback(*result_tuple(result, host))

        ref = self._make_id(client, ref_id)

        t = CallThread(self.call_client_retry, my_callback,
                       client, meth_name, ref, bin_file, owner, description, count)
        t.start()        

    def _get_node_list(self, proc, owner, callback):
        try:
            status = "OK"
            hosts = self.get_hosts(self.resource, self.subtype)
            data = []
            for h in hosts:
                s, n = proc(h, [], owner)
                if s == "OK":
                    data.extend(n)
        except Exception, e:
            status = e
            data = None
        if callback:
            callback(status, data)
        else:
            return (status, data)

    def _query_ids(self, host, ids, owner, callback, meth_name):
        # Allow query results to be constrained by owner
        def filter_results(results):
            if type(owner) not in (list, tuple):
                o = [owner]
            else:
                o = owner
            res = []
            for r in results:
                if r.owner in o:
                    res.append(r)
            return res

        if owner is None:
            return self._operate_on_ids(host, ids, callback, meth_name)
        else:
            return self._operate_on_ids(host, ids, callback, meth_name, 
                                        filter_results)


    def _operate_on_ids(self, host, ids, callback, meth_name, filter_results=None):
        if callback:
            assert callable(callback)

        if type(ids) not in (list, tuple):
            ids = (ids)

        client = self.client_pool.get_object()
        self._setup_client(client,
                           self.servers,
                           host,
                           meth_name)

        refs = self._make_id_list(client, ids)

        def result_tuple(result, host):
            data = None
            result = self._pretty_result(result, host)
            if isinstance(result, Exception):
                status = result
            else:
                status = _AviaryCommon._get_status(result.status)
                if status == "OK" and hasattr(result, "results"):
                    data = result.results
                    if filter_results:
                        data = filter_results(data)
            return (status, data)

        if callback:
            def my_callback(result):
                self.client_pool.return_object(client)
                callback(*result_tuple(result, host))

            t = CallThread(self.call_client_retry, my_callback,
                           client, meth_name, refs)
            t.start()
        else:
            def my_process_results(result):
                return result_tuple(result, host)

            res = self._call_sync(self.call_client_retry, my_process_results,
                                  client, meth_name, refs)
            self.client_pool.return_object(client)
            return res;

class AviaryOperations(Catalog):
    def __init__(self, name, datadir, 
                 locator_uri="", 
                 job_servers=True,
                 query_servers=True,
                 hadoop_servers=True,
                 key="", cert="", root_cert="", domain_verify=True):

        super(AviaryOperations, self).__init__()

        self.name = name
        if locator_uri:
            self.locator = AviaryLocator(datadir, locator_uri,
                                         key, cert, root_cert, domain_verify)
        else:
            self.locator = None

        transports = TransportFactory(key, cert, root_cert, domain_verify)
        transports.log_details(log, "AviaryOperations")

        # Add methods for the servers requested.
        # The __getattr__ method in Catalog will find them.
        # If locator has not been set, servers values should be strings
        self.mechs = []
        if job_servers:
            self.mechs.append(_AviaryJobMethods(self.locator, 
                                                transports, 
                                                datadir, job_servers))
            
        if query_servers:
            self.mechs.append(_AviaryQueryMethods(self.locator, 
                                                  transports, 
                                                  datadir, query_servers))

        if hadoop_servers:
            self.mechs.append(_AviaryHadoopMethods(self.locator,
                                                   transports, 
                                                   datadir, hadoop_servers))
        self.add_mechanisms(self.mechs)

    def get_hosts(self, resource, subtype=""):
        hosts = set()                                 
        for m in self.mechs:                
            hosts.update(m.get_hosts(resource, subtype))
        return list(hosts)
                        
                
            

